import logging
import psycopg2
import sqlite3
import sys

from argparse import ArgumentParser
from .helpers import get_connection


def main():
    p = ArgumentParser()
    p.add_argument("db", help="Database file (.db) or configuration (.ini)")
    p.add_argument("-l", "--limit", help="Max number of messages to log about rows", default="10")
    args = p.parse_args()

    # Set up logger
    logger = logging.getLogger()
    ch = logging.StreamHandler()
    formatter = logging.Formatter("%(levelname)s: %(message)s")
    ch.setFormatter(formatter)
    logger.setLevel(logging.WARNING)
    ch.setLevel(logging.WARNING)
    logger.addHandler(ch)

    # Parse limit
    lim = args.limit
    try:
        limit = int(lim)
    except ValueError:
        if lim.lower() == "none":
            limit = None
        else:
            logger.error("Invalid --limit option: " + lim)
            sys.exit(1)

    conn = get_connection(args.db)
    check(conn, limit=limit)


def check(conn, limit=10):
    logger = logging.getLogger()
    cur = conn.cursor()
    if isinstance(conn, sqlite3.Connection):
        cur.execute("SELECT name FROM sqlite_master WHERE type='table';")
    else:
        cur.execute(
            "SELECT table_name FROM information_schema.tables WHERE table_schema = 'public'"
        )
    tables = [x[0] for x in cur.fetchall()]

    if "prefix" not in tables:
        logger.error("missing 'prefix' table")
        prefix_ok = False
    else:
        prefix_ok = check_prefix(cur)

    statements_ok = True
    if "statements" not in tables:
        logger.error("missing 'statements' table")
        statements_ok = False
    elif prefix_ok:
        statements_ok = check_statements(cur, limit=limit)

    if not statements_ok or not prefix_ok:
        sys.exit(1)


def check_prefix(cur):
    """Check the structure of the prefix table. It must have the columns 'prefix' and 'base'."""
    logger = logging.getLogger()

    # Check for required columns
    if isinstance(cur, sqlite3.Cursor):
        cur.execute("PRAGMA table_info(prefix)")
        columns = {x[1]: x[2] for x in cur.fetchall()}
    else:
        columns = get_postgres_columns(cur, "prefix")
    missing = []
    bad_type = []
    for col in ["prefix", "base"]:
        coltype = columns.get(col)
        if not coltype:
            missing.append(col)
        elif coltype != "TEXT":
            bad_type.append(col)
    if missing:
        logger.error("'prefix' is missing column(s): " + ", ".join(missing))
        return False
    if bad_type:
        logger.error("'prefix' column(s) do not have type 'TEXT': " + ", ".join(bad_type))
        return False

    # Check for required prefixes
    missing_prefixes = []
    for prefix in ["owl", "rdf", "rdfs"]:
        cur.execute(f"SELECT * FROM prefix WHERE prefix = '{prefix}'")
        res = cur.fetchone()
        if not res:
            missing_prefixes.append(prefix)
    if missing_prefixes:
        logger.error("'prefix' is missing required prefixes: " + ", ".join(missing_prefixes))

    return True


def check_statements(cur, limit=10):
    """Check the structure of the statements table then check the values of the columns."""
    logger = logging.getLogger()
    statements_ok = True

    # First check the structure
    if isinstance(cur, sqlite3.Cursor):
        cur.execute("PRAGMA table_info(statements)")
        columns = {x[1]: x[2] for x in cur.fetchall()}
    else:
        columns = get_postgres_columns(cur, "statements")
    missing = []
    bad_type = []
    for col in [
        "stanza",
        "subject",
        "predicate",
        "object",
        "value",
        "datatype",
        "language",
    ]:
        coltype = columns.get(col)
        if not coltype:
            missing.append(col)
        elif coltype != "TEXT":
            bad_type.append(col)
    if missing:
        logger.error("'statements' is missing column(s): " + ", ".join(missing))
        return False
    if bad_type:
        logger.error("'statements' column(s) do not have type 'TEXT': " + ", ".join(bad_type))
        return False

    # Check for an index on the stanza column, warn if missing (do not fail)
    has_stanza_idx = False
    if isinstance(cur, sqlite3.Cursor):
        cur.execute("PRAGMA index_list(statements)")
        for row in cur.fetchall():
            index = row[1]
            cur.execute(f"PRAGMA index_info({index})")
            col = cur.fetchone()[2]
            if col == "stanza":
                has_stanza_idx = True
                break
    else:
        cur.execute(
            """SELECT a.attname AS column_name
               FROM pg_class t, pg_class i, pg_index ix, pg_attribute a
               WHERE
                t.oid = ix.indrelid
                and i.oid = ix.indexrelid
                and a.attrelid = t.oid
                and a.attnum = ANY(ix.indkey)
                and t.relkind = 'r'
                and t.relname = 'statements';"""
        )
        for row in cur.fetchall():
            if row[0] == "stanza":
                has_stanza_idx = True
                break
    if not has_stanza_idx:
        logger.warning("missing index on 'stanza' column")

    # Check that no row has both an object and a value
    message_count = 0
    cur.execute("SELECT * FROM statements WHERE object IS NOT NULL AND value IS NOT NULL")
    invalid = len(cur.fetchall())
    if invalid:
        logger.error(
            f"{invalid} rows where both 'object' and 'value' have values (one must be NULL)"
        )
        message_count += 1

    # Get prefixes to check against
    cur.execute("SELECT prefix, base FROM prefix")
    prefixes = {x[0]: x[1] for x in cur.fetchall()}

    # Check subjects, stanzas, predicates, and objects
    for col in ["stanza", "subject", "predicate", "object"]:
        if limit and message_count >= limit:
            # Do not exceed the limit of messages
            break
        cur.execute(f"SELECT {col} FROM statements")
        for row in cur.fetchall():
            value = row[0]
            if value is None:
                if col != "object":
                    # Object can be null when there is a value
                    logger.error(f"{col} cannot be NULL")
                    message_count += 1
                    statements_ok = False
                continue
            if value.startswith("<") and value.endswith(">"):
                iri = value.lstrip("<").rstrip(">")
                for prefix, base in prefixes.items():
                    if iri.startswith(base):
                        logger.warning(f"{col} '{value}' can use prefix '{prefix}'")
                        message_count += 1
                continue
            if ":" not in value:
                logger.error(f"{col} '{value}' is not a valid CURIE")
                message_count += 1
                statements_ok = False
                continue
            if value.startswith("_:"):
                if col == "predicate":
                    # Predicate should never be blank node, everything else is OK
                    logger.error(f"{col} '{value}' should be a named entity")
                    message_count += 1
                    statements_ok = False
                continue
            prefix = value.split(":")[0]
            if prefix not in prefixes:
                logger.error(f"{col} '{value}' does not have a valid prefix")
                message_count += 1
                statements_ok = False
    return statements_ok


def get_postgres_columns(cur, table):
    """Get a dictionary of column name to its type from a PostgreSQL database."""
    cur.execute(f"SELECT * FROM {table} LIMIT 0")
    columns = {}
    for desc in cur.description:
        type_oid = desc[1]
        cur.execute(f"SELECT typname FROM pg_type WHERE oid = {type_oid}")
        res = cur.fetchone()
        if res:
            typename = res[0].upper()
        else:
            typename = None
        columns[desc[0]] = typename
    return columns


if __name__ == "__main__":
    main()



import logging
import psycopg2
import re
import sqlite3

from configparser import ConfigParser
from rdflib import Graph


def add_labels(cur):
    """Create a temporary labels table. If a term does not have a label, the label is the ID."""
    # Create a tmp labels table
    cur.execute("CREATE TABLE tmp_labels(term TEXT PRIMARY KEY, label TEXT)")
    if isinstance(cur, sqlite3.Cursor):
        # Add all terms with label
        cur.execute(
            """INSERT OR IGNORE INTO tmp_labels SELECT subject, value
               FROM statements WHERE predicate = 'rdfs:label'"""
        )
        # Update remaining with their ID as their label
        cur.execute(
            "INSERT OR IGNORE INTO tmp_labels SELECT DISTINCT subject, subject FROM statements"
        )
        cur.execute(
            "INSERT OR IGNORE INTO tmp_labels SELECT DISTINCT predicate, predicate FROM statements"
        )
    else:
        # Do the same for a psycopg2 Cursor
        cur.execute(
            """INSERT INTO tmp_labels
               SELECT subject, value FROM statements WHERE predicate = 'rdfs:label'
               ON CONFLICT (term) DO NOTHING"""
        )
        cur.execute(
            """INSERT INTO tmp_labels
               SELECT DISTINCT subject, subject FROM statements
               ON CONFLICT (term) DO NOTHING"""
        )
        cur.execute(
            """INSERT INTO tmp_labels
               SELECT DISTINCT predicate, predicate FROM statements
               ON CONFLICT (term) DO NOTHING"""
        )


def escape(curie):
    """Escape illegal characters in the local ID portion of a CURIE"""
    prefix = curie.split(":")[0]
    local_id = curie.split(":")[1]
    local_id_fixed = re.sub(r"(?<!\\)([~!$&'()*+,;=/?#@%])", r"\\\1", local_id)
    return f"{prefix}:{local_id_fixed}"


def escape_qnames(cur, table):
    """Update CURIEs with illegal QName characters in the local ID by escaping those characters."""
    for keyword in ["stanza", "subject", "predicate", "object"]:
        cur.execute(
            f"""SELECT DISTINCT {keyword} FROM {table}
                WHERE {keyword} NOT LIKE '<%>' AND {keyword} NOT LIKE '_:%'"""
        )
        for row in cur.fetchall():
            curie = row[0]
            escaped = escape(curie)
            if curie != escaped:
                cur.execute(
                    f"UPDATE {table} SET {keyword} = '{escaped}' WHERE {keyword} = '{curie}'"
                )


def get_ancestors_capped(cur, top_terms, ancestors, term_id):
    """Return a set of ancestors for a given term ID, until a term in the top_terms is reached,
    or a top-level term is reached (below owl:Thing).

    :param cur: database Cursor object to query
    :param top_terms: set of top terms to stop at
    :param ancestors: set to collect ancestors in
    :param term_id: term ID to get the ancestors of"""
    cur.execute(
        f"""SELECT DISTINCT object FROM statements
            WHERE stanza = '{term_id}' AND predicate IN ('rdfs:subClassOf', 'rdfs:subPropertyOf')
              AND object NOT LIKE '_:%'"""
    )
    res = cur.fetchall()
    if not res:
        # No parents, we've hit the very top-level
        ancestors.add(term_id)
        return
    for r in res:
        parent = r[0]
        if parent == "owl:Thing" or (top_terms and r[0] in top_terms):
            continue
        ancestors.add(r[0])
        get_ancestors_capped(cur, top_terms, ancestors, parent)


def get_ancestors(cur, term_id):
    """Return a set of ancestors for a given term ID, all the way to the top-level (below owl:Thing)

    :param cur: database Cursor object to query
    :param term_id: term ID to get the ancestors of"""
    cur.execute(
        f"""WITH RECURSIVE ancestors(node) AS (
                VALUES ('{term_id}')
                UNION
                 SELECT object AS node
                FROM statements
                WHERE predicate IN ('rdfs:subClassOf', 'rdfs:subPropertyOf')
                  AND object = '{term_id}'
                UNION
                SELECT object AS node
                FROM statements, ancestors
                WHERE ancestors.node = statements.stanza
                  AND statements.predicate IN ('rdfs:subClassOf', 'rdfs:subPropertyOf')
                  AND statements.object NOT LIKE '_:%'
            )
            SELECT * FROM ancestors""",
    )
    return set([x[0] for x in cur.fetchall()])


def get_bottom_descendants(cur, term_id, descendants):
    """Get all bottom-level descendants for a given term with no intermediates. The bottom-level
    terms are those that are not ever used as the object of an rdfs:subClassOf statement.

    :param cur: database Cursor object to query
    :param term_id: term ID to get the bottom descendants of
    :param descendants: a set to add descendants to
    """
    cur.execute(
        f"""SELECT DISTINCT stanza FROM statements
            WHERE object = '{term_id}' AND predicate IN ('rdfs:subClassOf', 'rdfs:subPropertyOf')"""
    )
    res = cur.fetchall()
    if not res:
        # No children - this is a bottom-level term
        descendants.add(term_id)
        return
    for r in res:
        get_bottom_descendants(cur, r[0], descendants)


def get_children(cur, term_id):
    """Return a set of children for a given term ID."""
    cur.execute(
        f"""SELECT DISTINCT stanza FROM statements
            WHERE predicate IN ('rdfs:subClassOf', 'rdfs:subPropertyOf')
              AND object = '{term_id}'""",
    )
    return set([x[0] for x in cur.fetchall()])


def get_connection(file):
    """Given a file ending in .db or .ini, create a database connection."""
    if file.endswith(".db"):
        # Always SQLite
        logging.info("Initializing SQLite connection")
        return sqlite3.connect(file)
    elif file.endswith(".ini"):
        # Always PostgreSQL (for now)
        config_parser = ConfigParser()
        config_parser.read(file)
        if config_parser.has_section("postgresql"):
            params = {}
            for param in config_parser.items("postgresql"):
                params[param[0]] = param[1]
        else:
            logging.error(
                "Unable to create database connection; missing [postgresql] section from " + file
            )
            return None
        logging.info("Initializing PostgreSQL connection")
        return psycopg2.connect(**params)
    logging.error(
        "Either a database file or a config file must be specified with a .db or .ini extension"
    )
    return None


def get_descendants(cur, term_id):
    """Return a set of descendants for a given term ID."""
    cur.execute(
        f"""WITH RECURSIVE descendants(node) AS (
                VALUES ('{term_id}')
                UNION
                 SELECT stanza AS node
                FROM statements
                WHERE predicate IN ('rdfs:subClassOf', 'rdfs:subPropertyOf')
                  AND stanza = '{term_id}'
                UNION
                SELECT stanza AS node
                FROM statements, descendants
                WHERE descendants.node = statements.object
                  AND statements.predicate IN ('rdfs:subClassOf', 'rdfs:subPropertyOf')
            )
            SELECT * FROM descendants""",
    )
    return set([x[0] for x in cur.fetchall()])


def get_ids(cur, id_or_labels):
    """Create a list of IDs from a list of IDs or labels."""
    ids = []
    for id_or_label in id_or_labels:
        cur.execute(f"SELECT term FROM tmp_labels WHERE label = '{id_or_label}'")
        res = cur.fetchone()
        if res:
            ids.append(res[0])
        else:
            # Make sure this exists as an ID
            cur.execute(f"SELECT label FROM tmp_labels WHERE term = '{id_or_label}'")
            res = cur.fetchone()
            if res:
                ids.append(id_or_label)
            else:
                logging.warning(f" '{id_or_label}' does not exist in database")
    return ids


def get_parents(cur, term_id):
    """Return a set of parents for a given term ID."""
    cur.execute(
        f"""SELECT DISTINCT object FROM statements
            WHERE stanza = '{term_id}' AND predicate IN ('rdfs:subClassOf', 'rdfs:subPropertyOf')
            AND object NOT LIKE '_:%'"""
    )
    return set([x[0] for x in cur.fetchall()])


def get_terms(term_list, terms_file):
    """Get a list of terms from a list and/or a file from args."""
    terms = term_list or []
    if terms_file:
        with open(terms_file, "r") as f:
            for line in f:
                if line.startswith("#"):
                    continue
                if not line.strip():
                    continue
                m = re.match(r"(.+)\s#.+", line)
                if m:
                    terms.append(m.group(1).strip())
                else:
                    terms.append(line.strip())
    return terms


def get_top_ancestors(cur, ancestors, term_id, top_terms=None):
    """Get the top-level ancestor or ancestors for a given term with no intermediates. The top-level
    terms are those with no rdfs:subClassOf statement, or direct children of owl:Thing. If top_terms
    is included, they may also be those terms in that list.

    :param cur: database Cursor object to query
    :param ancestors: a set to add ancestors to
    :param term_id: term ID to get the top ancestor of
    :param top_terms: a list of top-level terms to stop at
                      (if an ancestor is in this set, it will be added and recursion will stop)
    """
    cur.execute(
        f"""SELECT DISTINCT object FROM statements
            WHERE stanza = '{term_id}' AND predicate IN ('rdfs:subClassOf', 'rdfs:subPropertyOf')
              AND object NOT LIKE '_:%'"""
    )
    res = cur.fetchall()
    if not res:
        ancestors.add(term_id)
        return
    for r in res:
        if r[0] == "owl:Thing":
            ancestors.add(term_id)
            break
        if top_terms and r[0] in top_terms:
            ancestors.add(r[0])
        else:
            get_top_ancestors(cur, ancestors, r[0], top_terms=top_terms)


def get_ttl(cur, table):
    """Get the given table as lines of Turtle (the lines are returned as a list)."""
    # Get ttl lines
    cur.execute(
        f"""WITH literal(value, escaped) AS (
              SELECT DISTINCT
                value,
                replace(replace(replace(value, '\\', '\\\\'), '"', '\\"'), '
            ', '\\n') AS escaped
              FROM {table}
            )
            SELECT
              '@prefix ' || prefix || ': <' || base || '> .'
            FROM prefix
            UNION ALL
            SELECT DISTINCT
               subject
            || ' '
            || predicate
            || ' '
            || coalesce(
                 object,
                 '"' || escaped || '"^^' || datatype,
                 '"' || escaped || '"@' || language,
                 '"' || escaped || '"'
               )
            || ' .'
            FROM {table} LEFT JOIN literal ON {table}.value = literal.value;"""
    )
    lines = []
    for row in cur.fetchall():
        line = row[0]
        if not line:
            continue
        # Replace newlines
        line = line.replace("\n", "\\n")
        lines.append(line)

    return "\n".join(lines)


def ttl_to_json(cur, ttl):
    # Create a Graph object from the TTL string
    graph = Graph()
    graph.parse(data=ttl, format="turtle")

    # Create the context with prefixes
    cur.execute("SELECT DISTINCT prefix, base FROM prefix;")
    context = {}
    for row in cur.fetchall():
        context[row[0]] = {"@id": row[1], "@type": "@id"}
    return graph.serialize(format="json-ld", context=context, indent=4).decode("utf-8")



import csv
import logging
import sqlite3
import sys

from argparse import ArgumentParser
from .helpers import (
    add_labels,
    escape_qnames,
    get_ancestors_capped,
    get_bottom_descendants,
    get_children,
    get_connection,
    get_descendants,
    get_ids,
    get_parents,
    get_terms,
    get_top_ancestors,
    get_ttl,
    ttl_to_json,
)

"""
Usage: python3 extract.py -d <sqlite-database> -t <curie> > <ttl-file>

Creates a TTL file containing the term, predicates, and ancestors. TTL is written to stdout.
You can include more than one `-t <curie>`/`--term <curie>`.

You may also specify multiple CURIEs to extract with `-T <file>`/`--terms <file>`
where the file contains a list of CURIEs to extract.

You may also specify which predicates you would like to include with
`-p <curie>`/`--predicate <curie>` or `-P <file>`/`--predicates <file>`
where the file contains a list of predicate CURIEs.

Finally, if you don't wish to include the ancestors of the term/terms,
include the `-n`/`--no-hierarchy` flag.

The sqlite-database must be created by RDFTab (https://github.com/ontodev/rdftab.rs)
and include 'statements' and 'prefixes' tables.

The CURIEs must use a prefix from the 'prefixes' table.
"""


def main():
    p = ArgumentParser()
    p.add_argument(
        "-d", "--database", required=True, help="Database file (.db) or configuration (.ini)"
    )
    p.add_argument("-t", "--term", action="append", help="CURIE or label of term to extract")
    p.add_argument(
        "-T", "--terms", help="File containing CURIES or labels of terms to extract",
    )
    p.add_argument(
        "-p", "--predicate", action="append", help="CURIE or label of predicate to include",
    )
    p.add_argument(
        "-P", "--predicates", help="File containing CURIEs or labels of predicates to include",
    )
    p.add_argument("-i", "--imports", help="TSV or CSV file containing import module details")
    p.add_argument("-c", "--config", help="Source configuration for imports")
    p.add_argument(
        "-I",
        "--intermediates",
        help="Included ancestor/descendant intermediates (default: all)",
        default="all",
    )
    p.add_argument(
        "-m", "--imported-from", help="IRI of source import ontology to annotate terms with"
    )
    p.add_argument(
        "-M",
        "--imported-from-property",
        help="ID of property to use for 'imported from' annotation (default: IAO:0000412)",
        default="IAO:0000412",
    )
    p.add_argument("-s", "--source", help="Ontology source to filter imports file")
    p.add_argument("-f", "--format", help="Output format (ttl or json)", default="ttl")
    p.add_argument(
        "-n",
        "--no-hierarchy",
        action="store_true",
        help="If provided, do not create any rdfs:subClassOf statements",
    )
    args = p.parse_args()
    sys.stdout.write(extract(args))


def extract(args):
    # Get required terms
    terms_list = get_terms(args.term, args.terms)
    terms = {}
    source = args.source
    if args.imports:
        terms = get_import_terms(args.imports, source=source)

    if not terms_list and not terms:
        logging.critical("One or more term(s) must be specified with --term, --terms, or --imports")
        sys.exit(1)

    for t in terms_list:
        if not args.no_hierarchy:
            terms[t] = {"Related": "ancestors"}
        else:
            terms[t] = {}

    predicates = get_terms(args.predicate, args.predicates)
    intermediates = args.intermediates
    imported_from = args.imported_from

    if args.config:
        # Get options from the config file based on the source
        if not source:
            logging.critical("A --source is required when using the --config option")
            sys.exit(1)
        config_path = args.config
        sep = "\t"
        if config_path.endswith(".csv"):
            sep = ","
        # Search for the source in the file and read in option
        found_source = False
        with open(config_path, "r") as f:
            reader = csv.DictReader(f, delimiter=sep)
            for row in reader:
                if row["Source"] == source:
                    found_source = True
                    intermediates = row.get("Intermediates", "all")
                    predicates_str = row.get("Predicates")
                    imported_from = row.get("IRI")
                    if predicates_str:
                        # Extend any existing command-line predicates
                        predicates.extend(predicates_str.split(" "))
                    break
        if not found_source:
            # No source with provided name found
            logging.critical(f"Source '{source}' does not exist in config file: " + config_path)
            sys.exit(1)

    # Get the database connection & extract terms
    conn = get_connection(args.database)
    try:
        return extract_terms(
            conn,
            terms,
            predicates,
            fmt=args.format,
            imported_from=imported_from,
            imported_from_property=args.imported_from_property,
            intermediates=intermediates,
            no_hierarchy=args.no_hierarchy,
        )
    finally:
        # Always remove temp tables before exiting
        clean(conn)


def clean(conn):
    cur = conn.cursor()
    cur.execute("DROP TABLE IF EXISTS tmp_labels")
    cur.execute("DROP TABLE IF EXISTS tmp_terms")
    cur.execute("DROP TABLE IF EXISTS tmp_predicates")
    cur.execute("DROP TABLE IF EXISTS tmp_extract")
    conn.commit()


def extract_terms(
    conn,
    terms,
    predicates,
    fmt="ttl",
    imported_from=None,
    imported_from_property="IAO:0000412",
    intermediates="all",
    no_hierarchy=False,
):
    """Extract terms from the ontology database and return the module as Turtle or JSON-LD."""
    if fmt.lower() not in ["ttl", "json-ld"]:
        raise Exception("Unknown format: " + fmt)

    intermediates = intermediates.lower()
    if intermediates not in ["all", "none"]:
        raise Exception("Unknown 'intermediates' option: " + intermediates)

    # Create a new table (extract) and copy the triples we care about
    # Then write the triples from that table to the output file
    cur = conn.cursor()

    # Pre-clean up
    clean(conn)

    # Create a temp labels table
    add_labels(cur)

    # First pass on terms, get all related entities
    ignore = []
    more_terms = set()
    for term_id, details in terms.items():
        # Confirm that this term exists
        cur.execute(f"SELECT * FROM statements WHERE stanza = '{term_id}' LIMIT 1")
        res = cur.fetchone()
        if not res:
            logging.warning(f"'{term_id}' does not exist in database")
            ignore.append(term_id)
            continue

        # Check for related entities & add them
        related = details.get("Related")
        if not related:
            continue
        related = related.strip().lower().split(" ")
        for r in related:
            if r == "ancestors":
                ancestors = set()
                if intermediates == "none":
                    # Find first ancestor/s that is/are either:
                    # - in the set of input terms
                    # - a top level term (below owl:Thing)
                    get_top_ancestors(cur, ancestors, term_id, top_terms=terms.keys())
                else:
                    # Otherwise get a set of ancestors, stopping at terms that are either:
                    # - in the set of input terms
                    # - a top level term (below owl:Thing)
                    get_ancestors_capped(cur, terms.keys(), ancestors, term_id)
                more_terms.update(ancestors)
            elif r == "children":
                # Just add the direct children
                more_terms.update(get_children(cur, term_id))
            elif r == "descendants":
                if intermediates == "none":
                    # Find all bottom-level descendants (do not have children)
                    descendants = set()
                    get_bottom_descendants(cur, term_id, descendants)
                    more_terms.update(descendants)
                else:
                    # Get a set of all descendants, including intermediates
                    more_terms.update(get_descendants(cur, term_id))
            elif r == "parents":
                # Just add the direct parents
                more_terms.update(get_parents(cur, term_id))
            else:
                # TODO: should this just warn and continue?
                raise Exception(f"unknown 'Related' keyword for '{term_id}': " + r)

    # Add those extra terms from related entities to our terms dict
    for mt in more_terms:
        if mt not in terms:
            # Don't worry about the parent ID because hierarchy will be maintained ...
            # ... based on the first ancestor in the full set of terms
            terms[mt] = {}

    predicate_ids = None
    if predicates:
        # Current predicates are IDs or labels - make sure we get all the IDs
        predicate_ids = get_ids(cur, predicates)

    # Create the terms table containing parent -> child relationships
    cur.execute("CREATE TABLE tmp_terms(child TEXT, parent TEXT)")
    for term_id in terms.keys():
        cur.execute(f"INSERT INTO tmp_terms VALUES ('{term_id}', NULL)")

    # Create tmp predicates table containing all predicates to include
    cur.execute("CREATE TABLE tmp_predicates(predicate TEXT PRIMARY KEY NOT NULL)")
    if predicate_ids:
        for predicate_id in predicate_ids:
            if isinstance(cur, sqlite3.Cursor):
                cur.execute(f"INSERT OR IGNORE INTO tmp_predicates VALUES ('{predicate_id}')")
            else:
                cur.execute(
                    f"""INSERT INTO tmp_predicates VALUES ('{predicate_id}')
                        ON CONFLICT (predicate) DO NOTHING"""
                )
    else:
        # Insert all predicates
        if isinstance(cur, sqlite3.Cursor):
            cur.execute(
                """INSERT OR IGNORE INTO tmp_predicates
                SELECT DISTINCT predicate
                FROM statements WHERE predicate NOT IN
                  ('rdfs:subClassOf', 'rdfs:subPropertyOf', 'rdf:type')"""
            )
        else:
            cur.execute(
                """INSERT INTO tmp_predicates
                SELECT DISTINCT predicate
                FROM statements WHERE predicate NOT IN
                  ('rdfs:subClassOf', 'rdfs:subPropertyOf', 'rdf:type')
                ON CONFLICT (predicate) DO NOTHING"""
            )

    # Add subclass/subproperty/type relationships to terms table
    for term_id, details in terms.items():
        # Check for overrides, regardless of no-hierarchy
        override_parent = details.get("Parent ID")
        if override_parent:
            # Just assert this as parent and don't worry about existing parent(s)
            cur.execute(f"INSERT INTO tmp_terms VALUES ('{term_id}', '{override_parent}')")
            continue
        if no_hierarchy:
            continue

        # Otherwise only add the parent if we want a hierarchy
        # Check for the first ancestor we can find with all terms considered "top level"
        # In many cases, this is just the direct parent
        parents = set()
        get_top_ancestors(cur, parents, term_id, top_terms=terms.keys())
        parents = parents.intersection(set(terms.keys()))
        if parents:
            # Maintain these relationships in the import module
            for p in parents:
                if p == term_id:
                    continue
                cur.execute(f"INSERT INTO tmp_terms VALUES ('{term_id}', '{p}')")

    # Create our extract table to hold the actual triples
    cur.execute(
        """CREATE TABLE tmp_extract(
             stanza TEXT,
             subject TEXT,
             predicate TEXT,
             object TEXT,
             value TEXT,
             datatype TEXT,
             language TEXT
           )"""
    )

    # Insert rdf:type declarations - only for OWL entities
    cur.execute(
        """INSERT INTO tmp_extract
        SELECT * FROM statements
        WHERE subject IN (SELECT DISTINCT child FROM tmp_terms)
          AND predicate = 'rdf:type'
          AND object IN
          ('owl:Class', 'owl:AnnotationProperty', 'owl:DataProperty', 'owl:ObjectProperty', 'owl:NamedIndividual')"""
    )

    # Insert subproperty statements for any property types
    cur.execute(
        """INSERT INTO tmp_extract (stanza, subject, predicate, object)
        SELECT DISTINCT child, child, 'rdfs:subPropertyOf', parent
        FROM tmp_terms WHERE parent IS NOT NULL AND child IN
          (SELECT subject FROM statements WHERE predicate = 'rdf:type'
           AND object IN ('owl:AnnotationProperty', 'owl:DataProperty', 'owl:ObjectProperty')
           AND subject NOT LIKE '_:%')"""
    )

    # Insert subclass statements for any class types
    cur.execute(
        """INSERT INTO tmp_extract (stanza, subject, predicate, object)
        SELECT DISTINCT child, child, 'rdfs:subClassOf', parent
        FROM tmp_terms WHERE parent IS NOT NULL AND child IN
          (SELECT subject FROM statements WHERE predicate = 'rdf:type'
           AND object = 'owl:Class' AND subject NOT LIKE '_:%')"""
    )

    # Everything else is an instance
    # TODO: or datatype?
    cur.execute(
        """INSERT INTO tmp_extract (stanza, subject, predicate, object)
        SELECT DISTINCT child, child, 'rdf:type', parent
        FROM tmp_terms WHERE parent IS NOT NULL AND child NOT IN
          (SELECT stanza from tmp_extract
           WHERE predicate IN ('rdfs:subClassOf', 'rdfs:subPropertyOf'))"""
    )

    # Insert literal annotations
    cur.execute(
        """INSERT INTO tmp_extract
        SELECT *
        FROM statements
        WHERE subject IN (SELECT DISTINCT child FROM tmp_terms)
          AND predicate IN (SELECT predicate FROM tmp_predicates)
          AND value IS NOT NULL"""
    )

    # Insert logical relationships (object must be in set of input terms)
    cur.execute(
        """INSERT INTO tmp_extract
        SELECT * FROM statements
        WHERE subject IN (SELECT DISTINCT child FROM tmp_terms)
          AND predicate IN (SELECT predicate FROM tmp_predicates)
          AND object IN (SELECT DISTINCT child FROM tmp_terms)"""
    )

    # Insert IRI annotations (object does not have to be in input terms)
    cur.execute(
        """INSERT INTO tmp_extract (stanza, subject, predicate, object)
        SELECT s1.stanza, s1.subject, s1.predicate, s1.object
        FROM statements s1
        JOIN statements s2 ON s1.predicate = s2.subject
        WHERE s1.subject IN (SELECT DISTINCT child FROM tmp_terms)
          AND s1.predicate IN (SELECT predicate FROM tmp_predicates)
          AND s2.object = 'owl:AnnotationProperty'
          AND s1.object IS NOT NULL"""
    )

    # Finally, if imported_from IRI is included, add this to add terms
    if imported_from:
        cur.execute(
            f"""INSERT INTO tmp_extract (stanza, subject, predicate, object)
            SELECT DISTINCT child, child, 'IAO:0000412', '<{imported_from}>' FROM tmp_terms"""
        )

    # Escape QNames
    escape_qnames(cur, "tmp_extract")

    ttl = get_ttl(cur, "tmp_extract")
    if fmt.lower() == "ttl":
        return ttl

    # Otherwise the format is JSON
    return ttl_to_json(cur, ttl)


def get_import_terms(import_file, source=None):
    """Get the terms and their details from the input file.

    :param import_file: path to file containing import details
    :param source: source ontology ID for terms to include
                   (if None or if input has no 'source' column, all are included)
    """
    terms = {}
    sep = "\t"
    if import_file.endswith(".csv"):
        sep = "\t"
    with open(import_file, "r") as f:
        reader = csv.DictReader(f, delimiter=sep)
        for row in reader:
            term_id = row.get("ID")
            if not term_id:
                continue
            if source and row.get("Source") != source:
                # If we have a source and this is not from that source, skip
                continue
            terms[term_id] = {"Parent ID": row.get("Parent ID"), "Related": row.get("Related")}
    return terms


if __name__ == "__main__":
    main()



import csv
import io
import logging
import re
import sys

from argparse import ArgumentParser
from collections import defaultdict

from .helpers import add_labels, get_connection, get_ids, get_terms
from .hiccup import render

"""
Usage: python3 -m gizmos.export -d <sqlite-database> -t <term-curie> > <output-file>

Creates a TSV output with the term details where the fields are the predicates.

The sqlite-database must be created by RDFTab (https://github.com/ontodev/rdftab.rs)
and include 'statements' and 'prefixes' tables.

You may specify multiple CURIEs to extract with `-T <file>`/`--terms <file>`
where the file contains a list of CURIEs to extract.

You may also specify which predicates you would like to include with
`-p <curie>`/`--predicate <curie>` or `-P <file>`/`--predicates <file>`
where the file contains a list of predicate CURIEs.

You can optionally specify a different format:
- CSV
- JSON
- HTML

When there is more than one value for a predicate, the values will be separated by a comma 
(unless you are writing to JSON, in which case it will be an array). 
If you want to split on a different character, use `-s <split>`

Finally, if writing to TSV, CSV, or HTML, you can also pass -n/--no-headers 
to exclude the table headers.
"""


def main():
    p = ArgumentParser()
    p.add_argument(
        "-d", "--database", required=True, help="Database file (.db) or configuration (.ini)"
    )
    p.add_argument("-t", "--term", action="append", help="CURIE or label of term to extract")
    p.add_argument(
        "-T", "--terms", help="File containing CURIES or labels of terms to extract",
    )
    p.add_argument(
        "-p", "--predicate", action="append", help="CURIE or label of predicate to include",
    )
    p.add_argument(
        "-P", "--predicates", help="File containing CURIEs or labels of predicates to include",
    )
    p.add_argument("-f", "--format", help="Output format (tsv, csv, html)", default="tsv")
    p.add_argument("-s", "--split", help="Character to split multiple values on", default="|")
    p.add_argument(
        "-c",
        "--contents-only",
        action="store_true",
        help="If provided with HTML format, render HTML without roots",
    )
    p.add_argument("-V", "--values", help="Default value format for cell values", default="IRI")
    p.add_argument("-w", "--where", help="SQL WHERE statement to include when selecting terms")
    p.add_argument(
        "-n",
        "--no-headers",
        action="store_true",
        help="If provided, do not include headers in output table",
    )
    args = p.parse_args()
    sys.stdout.write(export(args))


def export(args):
    """Wrapper for export_terms."""
    terms = get_terms(args.term, args.terms)
    predicates = get_terms(args.predicate, args.predicates)
    conn = get_connection(args.database)
    try:
        return export_terms(
            conn,
            terms,
            predicates,
            args.format,
            default_value_format=args.values,
            standalone=not args.contents_only,
            split=args.split,
            no_headers=args.no_headers,
            where=args.where,
        )
    finally:
        # Post clean-up
        cur = conn.cursor()
        cur.execute("DROP TABLE IF EXISTS tmp_labels")
        conn.commit()


def get_html_value(value_format, predicate_id, vo):
    """Return a hiccup-style HTML href or simple string for a value or object dictionary based on
    the value format. The href will only be returned if the dictionary has an 'iri' key."""
    if "value" in vo:
        return ["p", {"property": predicate_id}, vo["value"]]
    elif value_format == "label":
        iri = vo.get("iri")
        text = vo.get("label") or vo["id"]
    elif value_format == "curie":
        iri = vo.get("iri")
        text = vo["id"]
    else:
        iri = vo.get("iri")
        text = iri
    if predicate_id not in ["CURIE", "IRI", "label"]:
        return ["p", ["a", {"property": predicate_id, "resource": vo["id"], "href": iri}, text]]
    if predicate_id == "label":
        return ["p", {"property": "rdfs:label"}, text]
    return ["a", {"href": iri}, text]


def get_iri(prefixes, term):
    """Get the IRI from a CURIE."""
    if term.startswith("<"):
        return term.lstrip("<").rstrip(">")
    prefix = term.split(":")[0]
    namespace = prefixes.get(prefix)
    if not namespace:
        raise Exception(f"Prefix '{prefix}' is not defined in prefix table")
    local_id = term.split(":")[1]
    return namespace + local_id


def get_objects(cur, prefixes, term, predicate_ids):
    """Get a dict of predicate label -> objects. The object will either be the term ID or label,
    when the label exists."""
    predicates_str = ", ".join(
        [f"'{x}'" for x in predicate_ids.keys() if x not in ["CURIE", "IRI", "label"]]
    )
    term_objects = defaultdict(list)
    cur.execute(
        f"""SELECT DISTINCT predicate, s.object AS object, l.label AS object_label
            FROM statements s JOIN tmp_labels l ON s.object = l.term
            WHERE s.subject = '{term}' AND s.predicate IN ({predicates_str})"""
    )
    for row in cur.fetchall():
        p = row[0]
        p_label = predicate_ids[p]
        if p_label not in term_objects:
            term_objects[p_label] = list()

        obj = row[1]
        if obj.startswith("_:"):
            # TODO - handle blank nodes
            continue
        obj_label = row[2]

        d = {"id": obj, "iri": get_iri(prefixes, term)}
        # Maybe add the label
        if obj != obj_label:
            d["label"] = obj_label
        term_objects[p_label].append(d)
    return term_objects


def get_predicate_ids(cur, id_or_labels=None):
    """Create a map of predicate label or full header (if the header has a value format) -> ID."""
    predicate_ids = {}
    if id_or_labels:
        for id_or_label in id_or_labels:
            m = re.match(r"(.+) \[.+]$", id_or_label)
            if m:
                id_or_label = m.group(1)
            if id_or_label in ["CURIE", "IRI", "label"]:
                predicate_ids[id_or_label] = id_or_label
                continue
            cur.execute(f"SELECT term FROM tmp_labels WHERE label = '{id_or_label}'")
            res = cur.fetchone()
            if res:
                predicate_ids[res[0]] = id_or_label
            else:
                # Make sure this exists as an ID
                cur.execute(f"SELECT label FROM tmp_labels WHERE term = '{id_or_label}'")
                res = cur.fetchone()
                if res:
                    predicate_ids[id_or_label] = id_or_label
                else:
                    logging.warning(f"'{id_or_label}' does not exist in database")
        return predicate_ids

    cur.execute(
        """SELECT DISTINCT s.predicate AS term, l.label AS label
           FROM statements s JOIN tmp_labels l ON s.predicate = l.term"""
    )
    for row in cur.fetchall():
        predicate_ids[row[0]] = row[1]
    if "rdf:type" in predicate_ids:
        del predicate_ids["rdf:type"]
    return predicate_ids


def get_string_value(value_format, vo):
    """Return a string from a value or object dictionary based on the value format."""
    if "value" in vo:
        return vo["value"]
    elif value_format == "label":
        # Label or CURIE (when no label)
        return vo.get("label") or vo["id"]
    elif value_format == "curie":
        # Always the CURIE
        return vo["id"]
    # IRI or CURIE (when no IRI, which shouldn't happen)
    return vo.get("iri") or vo["id"]


def get_term_details(cur, prefixes, term, predicate_ids):
    """Get a dict of predicate label -> object or value."""
    term_details = {}

    # Handle special cases
    cur.execute(f"SELECT label FROM tmp_labels WHERE term = '{term}'")
    res = cur.fetchone()
    base_dict = {"id": term, "iri": get_iri(prefixes, term)}
    if res:
        base_dict["label"] = res[0]
    if "CURIE" in predicate_ids:
        term_details["CURIE"] = base_dict
    if "IRI" in predicate_ids:
        term_details["IRI"] = base_dict
    if "label" in predicate_ids:
        term_details["label"] = base_dict

    # Get all details
    term_details.update(get_values(cur, term, predicate_ids))
    term_details.update(get_objects(cur, prefixes, term, predicate_ids))

    # Format predicates with multiple values - a single value should not be an array
    term_details_fixed = {}
    for predicate, values in term_details.items():
        if len(values) == 1:
            term_details_fixed[predicate] = values[0]
        else:
            term_details_fixed[predicate] = values
    return term_details_fixed


def get_values(cur, term, predicate_ids):
    """Get a dict of predicate label -> literal values."""
    predicates_str = ", ".join(
        [f"'{x}'" for x in predicate_ids.keys() if x not in ["CURIE", "IRI", "label"]]
    )
    term_values = defaultdict(list)
    cur.execute(
        f"""SELECT DISTINCT predicate, value FROM statements s
            WHERE subject = '{term}' AND predicate IN ({predicates_str}) AND value IS NOT NULL"""
    )
    for row in cur.fetchall():
        p = row[0]
        p_label = predicate_ids[p]
        value = row[1]
        if value:
            if p_label not in term_values:
                term_values[p_label] = list()
            term_values[p_label].append({"value": value})
    return term_values


def render_html(prefixes, value_formats, predicate_ids, details, standalone=True, no_headers=False):
    """Render an HTML table."""
    predicate_labels = {v: k for k, v in predicate_ids.items()}
    # Create the prefix element
    pref_strs = []
    for prefix, base in prefixes.items():
        pref_strs.append(f"{prefix}: {base}")
    pref_str = "\n".join(pref_strs)
    table = ["table", {"class": "table table-striped", "prefix": pref_str}]

    # Get headers - in order
    headers = []
    for k in value_formats.keys():
        headers.append(k)

    if not no_headers:
        # Table headers
        thead = ["thead"]
        tr = ["tr"]
        for h in headers:
            tr.append(["th", h])
        thead.append(tr)
        table.append(thead)

    # Table body
    tbody = ["tbody"]
    for term, detail in details.items():
        tr = ["tr", {"resource": term}]
        for h in headers:
            m = re.match(r"(.+) \[.+]", h)
            if m:
                pred_label = m.group(1)
            else:
                pred_label = h

            predicate_id = predicate_labels.get(pred_label)
            value_format = value_formats.get(h)
            vo_list = detail.get(pred_label)
            if not vo_list:
                tr.append(["td"])
                continue
            if isinstance(vo_list, list):
                items = []
                for vo in vo_list:
                    items.append(get_html_value(value_format, predicate_id, vo))
                ele = ["td"] + items
                tr.append(ele)
            else:
                display = get_html_value(value_format, predicate_id, vo_list)
                if isinstance(display, str):
                    if predicate_id == "label":
                        predicate_id = "rdfs:label"
                    display = ["p", {"property": predicate_id}, display]
                tr.append(["td", display])
        tbody.append(tr)
    table.append(tbody)

    # Render full HTML
    if standalone:
        # HTML Headers & CSS
        head = [
            "head",
            ["meta", {"charset": "utf-8"}],
            [
                "meta",
                {
                    "name": "viewport",
                    "content": "width=device-width, initial-scale=1, shrink-to-fit=no",
                },
            ],
            [
                "link",
                {
                    "rel": "stylesheet",
                    "href": "https://stackpath.bootstrapcdn.com/bootstrap/4.5.0/css/bootstrap.min.css",
                    "crossorigin": "anonymous",
                },
            ],
        ]
        html = ["html", head, ["body", table]]
    else:
        html = table
    return render(None, html)


def render_output(
    prefixes,
    value_formats,
    predicate_ids,
    details,
    fmt,
    split="|",
    standalone=True,
    no_headers=False,
):
    """Render the string output based on the format."""
    if fmt == "tsv":
        return render_table(value_formats, details, "\t", split=split, no_headers=no_headers)
    elif fmt == "csv":
        return render_table(value_formats, details, ",", split=split, no_headers=no_headers)
    elif fmt == "html":
        return render_html(
            prefixes,
            value_formats,
            predicate_ids,
            details,
            standalone=standalone,
            no_headers=no_headers,
        )
    else:
        raise Exception("Invalid format: " + fmt)


def render_table(value_formats, details, separator, split="|", no_headers=False):
    """Render a TSV or CSV table."""
    # First fix the output to be writable by DictWriter
    rows = []
    for d in details.values():
        row = {}
        for header, value_format in value_formats.items():
            m = re.match(r"(.+) \[.+]", header)
            if m:
                pred_label = m.group(1)
            else:
                pred_label = header
            value = d.get(pred_label)
            if not value:
                continue
            if isinstance(value, list):
                items = []
                for itm in value:
                    items.append(get_string_value(value_format, itm))
                value = split.join(items)
                row[header] = value
            else:
                row[header] = get_string_value(value_format, value)
        rows.append(row)

    # Then get headers - in order
    headers = []
    for k in value_formats.keys():
        headers.append(k)

    # Finally write to string
    output = io.StringIO()
    writer = csv.DictWriter(output, fieldnames=headers, delimiter=separator, lineterminator="\n")
    if not no_headers:
        writer.writeheader()
    writer.writerows(rows)
    return output.getvalue()


def export_terms(
    conn,
    terms,
    predicates,
    fmt,
    split="|",
    standalone=True,
    default_value_format="IRI",
    no_headers=False,
    where=None,
):
    """Retrieve details for given terms and render in the given format."""

    # Validate default format
    if default_value_format not in ["CURIE", "IRI", "label"]:
        raise Exception(
            f"The default value format ('{default_value_format}') must be one of: CURIE, IRI, label"
        )
    # Validate output format
    if fmt.lower() not in ["tsv", "csv", "html"]:
        raise Exception(f"Output format '{fmt}' must be one of: tsv, csv, html")

    details = {}
    # Check if writing to JSON - the dict output will be different

    cur = conn.cursor()

    # Create a tmp labels table
    cur.execute("DROP TABLE IF EXISTS tmp_labels")
    conn.commit()
    add_labels(cur)

    if terms:
        term_ids = get_ids(cur, terms)
    else:
        term_ids = []
        if where:
            # Use provided query filter to select terms
            query = "SELECT DISTINCT stanza FROM statements WHERE " + where
        else:
            # Get all, excluding blank nodes
            query = "SELECT DISTINCT stanza FROM statements WHERE stanza NOT LIKE '_:%'"
        cur.execute(query)
        for row in cur.fetchall():
            term_ids.append(row[0])

    if not predicates:
        # Get all predicates if not provided
        predicate_ids = {default_value_format: default_value_format}
        value_formats = {default_value_format: default_value_format.lower()}
        predicate_ids.update(get_predicate_ids(cur))
        predicate_id_str = ", ".join([f"'{x}'" for x in predicate_ids.keys()])
        cur.execute(f"SELECT DISTINCT label FROM tmp_labels WHERE term IN ({predicate_id_str})")
        for row in cur.fetchall():
            value_formats[row[0]] = default_value_format.lower()

    else:
        # Current predicates are IDs or labels - make sure we get all the IDs
        predicate_ids = get_predicate_ids(cur, predicates)
        value_formats = {}
        for p in predicates:
            if p in ["CURIE", "IRI", "label"]:
                value_format = p.lower()
            else:
                value_format = default_value_format.lower()
                m = re.match(r".+ \[(.+)]$", p)
                if m:
                    value_format = m.group(1).lower()
            value_formats[p] = value_format

    # Get prefixes
    prefixes = {}
    cur.execute(f"SELECT DISTINCT prefix, base FROM prefix")
    for row in cur.fetchall():
        prefixes[row[0]] = row[1]

    # Get the term details
    for term in term_ids:
        term_details = get_term_details(cur, prefixes, term, predicate_ids)
        details[term] = term_details

    return render_output(
        prefixes,
        value_formats,
        predicate_ids,
        details,
        fmt,
        split=split,
        standalone=standalone,
        no_headers=no_headers,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import logging
import os
import sys

from argparse import ArgumentParser
from collections import defaultdict
from gizmos.hiccup import render
from .helpers import get_connection

"""
Usage: python3 tree.py <sqlite-database> <term-curie> > <html-file>

Creates an HTML page containing the tree structure of the term & its annotations.
HTML is written to stdout.

The sqlite-database must be created by RDFTab (https://github.com/ontodev/rdftab.rs)
and include 'statements' and 'prefixes' tables.

The term-curie must use a prefix from the 'prefixes' table.
"""

LOGGER = logging.getLogger("main")
logging.basicConfig(
    level=logging.INFO, format="%(levelname)s - %(asctime)s - %(name)s - %(message)s"
)

# Plus sign to show a node has children
PLUS = [
    "svg",
    {"width": "14", "height": "14", "fill": "#808080", "viewBox": "0 0 16 16"},
    [
        "path",
        {
            "fill-rule": "evenodd",
            "d": "M8 15A7 7 0 1 0 8 1a7 7 0 0 0 0 14zm0 1A8 8 0 1 0 8 0a8 8 0 0 0 0 16z",
        },
    ],
    [
        "path",
        {
            "fill-rule": "evenodd",
            "d": "M8 4a.5.5 0 0 1 .5.5v3h3a.5.5 0 0 1 0 1h-3v3a.5.5 0 0 1-1 0v-3h-3a.5.5 0 0 1 "
            + "0-1h3v-3A.5.5 0 0 1 8 4z",
        },
    ],
]

# Top levels to display in tree
TOP_LEVELS = {
    "ontology": "Ontology",
    "owl:Class": "Class",
    "owl:AnnotationProperty": "Annotation Property",
    "owl:DataProperty": "Data Property",
    "owl:ObjectProperty": "Object Property",
    "owl:Individual": "Individual",
    "rdfs:Datatype": "Datatype",
}

# Stylesheets & JS scripts
bootstrap_css = "https://stackpath.bootstrapcdn.com/bootstrap/4.5.0/css/bootstrap.min.css"
bootstrap_js = "https://stackpath.bootstrapcdn.com/bootstrap/4.4.1/js/bootstrap.min.js"
popper_js = "https://cdn.jsdelivr.net/npm/popper.js@1.16.0/dist/umd/popper.min.js"
typeahead_js = "https://cdnjs.cloudflare.com/ajax/libs/typeahead.js/0.11.1/typeahead.bundle.min.js"


def main():
    p = ArgumentParser("tree.py", description="create an HTML page to display an ontology term")
    p.add_argument("db", help="Database file (.db) or configuration (.ini)")
    p.add_argument("term", help="CURIE of ontology term to display", nargs="?")
    p.add_argument("-t", "--title", help="Optional tree title")
    p.add_argument("-p", "--predicate", action="append", help="CURIE of predicate to include")
    p.add_argument("-P", "--predicates", help="File containing CURIEs of predicates to include")
    p.add_argument(
        "-d",
        "--include-db",
        help="If provided, include 'db' param in query string",
        action="store_true",
    )
    p.add_argument("-H", "--href", help="Format string to convert CURIEs to tree links")
    p.add_argument(
        "-s", "--include-search", help="If provided, include a search bar", action="store_true"
    )
    p.add_argument(
        "-c",
        "--contents-only",
        action="store_true",
        help="If provided, render HTML without the roots",
    )
    args = p.parse_args()

    # Maybe get predicates to include
    predicate_ids = args.predicate or []
    if args.predicates:
        with open(args.predicates, "r") as f:
            predicate_ids.extend([x.strip() for x in f.readlines()])

    if args.href:
        href = args.href
        if "{curie}" not in href:
            raise RuntimeError("The --href argument must contain '{curie}'")
    else:
        href = "?id={curie}"
        if args.include_db:
            href += "&db={db}"

    treename = os.path.splitext(os.path.basename(args.db))[0]
    conn = get_connection(args.db)

    # Run tree and write HTML to stdout
    sys.stdout.write(
        tree(
            conn,
            treename,
            args.term,
            title=args.title,
            href=href,
            predicate_ids=predicate_ids,
            include_search=args.include_search,
            standalone=not args.contents_only,
        )
    )


def tree(
    conn,
    treename,
    term,
    href="?id={curie}",
    title=None,
    predicate_ids=None,
    include_search=False,
    standalone=True,
):
    cur = conn.cursor()
    return build_tree(
        cur,
        treename,
        term_id=term,
        title=title,
        href=href,
        predicate_ids=predicate_ids,
        include_search=include_search,
        standalone=standalone,
    )


def annotations2rdfa(treename, data, predicate_ids, term_id, stanza, href="?term={curie}"):
    """Create a hiccup-style vector for the annotation on a term."""
    # The subjects in the stanza that are of type owl:Axiom:
    annotation_bnodes = set()
    for row in stanza:
        if row["predicate"] == "owl:annotatedSource":
            annotation_bnodes.add(row["subject"])

    # Annotations, etc. on the right-hand side for the subjects contained in
    # annotation_bnodes:
    annotations = defaultdict(dict)
    for row in stanza:
        # subject is the blank node, _:...
        subject = row["subject"]
        if subject not in annotation_bnodes:
            continue

        if subject not in annotations:
            annotations[subject] = {}

        predicate = row["predicate"]
        obj = row["object"]
        value = row["value"]

        if predicate not in [
            "owl:annotatedSource",
            "owl:annotatedTarget",
            "owl:annotatedProperty",
            "rdf:type",
        ]:
            # This is the actual axiom that we care about and contains display value
            annotations[subject]["predicate"] = predicate
            if obj:
                annotations[subject]["object"] = obj
            if value:
                annotations[subject]["value"] = value
            annotations[subject]["annotation"] = row

        if predicate == "owl:annotatedSource":
            annotations[subject]["source"] = obj

        elif predicate == "owl:annotatedProperty":
            annotations[subject]["target_predicate"] = obj

        elif predicate == "owl:annotatedTarget":
            if obj:
                annotations[subject]["target_object"] = obj
            if value:
                annotations[subject]["target_value"] = value

    spv2annotation = {}
    for bnode, details in annotations.items():
        source = details["source"]
        target_predicate = details["target_predicate"]
        target = details.get("target_object", None) or details.get("target_value", None)

        if source in spv2annotation:
            # list of predicate -> values on this target (combo of predicate + value)
            pred2val = spv2annotation[source]
        else:
            pred2val = {}

        if target_predicate in pred2val:
            annotated_values = pred2val[target_predicate]
        else:
            annotated_values = {}

        if target in annotated_values:
            ax_annotations = annotated_values[target]
        else:
            ax_annotations = {}

        # predicate of the annotation
        ann_predicate = details["predicate"]
        if ann_predicate in ax_annotations:
            # values of the annotation
            anns = ax_annotations[ann_predicate]
        else:
            anns = []
        anns.append(details["annotation"])

        ax_annotations[ann_predicate] = anns
        annotated_values[target] = ax_annotations
        pred2val[target_predicate] = annotated_values
        spv2annotation[source] = pred2val

    # The initial hiccup, which will be filled in later:
    items = ["ul", {"id": "annotations", "class": "col-md"}]
    labels = data["labels"]

    # s2 maps the predicates of the given term to their corresponding rows (there can be more than
    # one row per predicate):
    s2 = defaultdict(list)
    for row in stanza:
        if row["subject"] == term_id:
            s2[row["predicate"]].append(row)
    pcs = list(s2.keys())

    # Loop through the rows of the stanza that correspond to the predicates of the given term:
    for predicate in predicate_ids:
        if predicate not in pcs:
            continue
        predicate_label = predicate
        if predicate.startswith("<"):
            predicate_label = predicate.lstrip("<").rstrip(">")
        anchor = [
            "a",
            {"href": href.format(curie=predicate, db=treename)},
            labels.get(predicate, predicate_label),
        ]
        # Initialise an empty list of "o"s, i.e., hiccup representations of objects:
        objs = []
        for row in s2[predicate]:
            # Convert the `data` map, that has entries for the tree and for a list of the labels
            # corresponding to all of the curies in the stanza, into a hiccup object `o`:
            o = ["li", row2o(stanza, data, row)]

            # Check for axiom annotations and create nested
            nest = build_nested(treename, data, labels, spv2annotation, term_id, row, [], href=href)
            if nest:
                o += nest

            # Append the `o` to the list of `os`:
            objs.append(o)
        if objs:
            items.append(["li", anchor, ["ul"] + objs])
    return items


def build_nested(treename, data, labels, spv2annotation, source, row, ele, href="?id={curie}"):
    """Build a nested hiccup list of axiom annotations."""
    predicate = row["predicate"]
    if source in spv2annotation:
        annotated_predicates = spv2annotation[source]
        if predicate in annotated_predicates:
            annotated_values = annotated_predicates[predicate]
            target = row.get("object", None) or row.get("value", None)
            if target in annotated_values:
                ax_annotations = annotated_values[target]
                for ann_predicate, ann_rows in ax_annotations.items():
                    # Build the nested list "anchor" (predicate)
                    anchor = [
                        "li",
                        [
                            "small",
                            [
                                "a",
                                {"href": href.format(curie=ann_predicate, db=treename)},
                                labels.get(ann_predicate, ann_predicate),
                            ],
                        ],
                    ]

                    # Collect the axiom annotation objects/values
                    ax_os = []
                    for ar in ann_rows:
                        ax_os.append(["li", ["small", row2o([], data, ar)]])
                        build_nested(
                            treename,
                            data,
                            labels,
                            spv2annotation,
                            ar["subject"],
                            ar,
                            ax_os,
                            href=href,
                        )
                    ele.append(["ul", anchor, ["ul"] + ax_os])
    return ele


def build_tree(
    cur,
    treename,
    term_id=None,
    title=None,
    href="?id={curie}",
    predicate_ids=None,
    include_search=False,
    standalone=True,
):
    """Create a hiccup-style HTML vector for the given terms.
    If there are no terms, create the HTML vector for all top-level classes."""
    # Get the prefixes
    cur.execute("SELECT * FROM prefix ORDER BY length(base) DESC")
    all_prefixes = [(x[0], x[1]) for x in cur.fetchall()]

    ps = set()
    body = []
    if not term_id:
        p, t = term2rdfa(
            cur, all_prefixes, treename, predicate_ids, "owl:Class", [], title=title, href=href
        )
        ps.update(p)
        body.append(t)

    # Maybe find a * in the IDs that represents all remaining predicates
    predicate_ids_split = None
    if predicate_ids and "*" in predicate_ids:
        before = []
        after = []
        found = False
        for pred in predicate_ids:
            if pred == "*":
                found = True
                continue
            if not found:
                before.append(pred)
            else:
                after.append(pred)
        predicate_ids_split = [before, after]

    # Run for given terms if terms have not yet been filled out
    if not body:
        if predicate_ids and predicate_ids_split:
            # If some IDs were provided with *, add the remaining predicates
            # These properties go in between the before & after defined in the split
            rem_predicate_ids = get_sorted_predicates(cur, exclude_ids=predicate_ids)

            # Separate before & after with the remaining properties
            predicate_ids = predicate_ids_split[0]
            predicate_ids.extend(rem_predicate_ids)
            predicate_ids.extend(predicate_ids_split[1])
        elif not predicate_ids:
            predicate_ids = get_sorted_predicates(cur)

        cur.execute(
            f"""SELECT stanza, subject, predicate, object, value, datatype, language
                FROM statements WHERE stanza = '{term_id}'"""
        )
        stanza = create_stanza(cur.fetchall())
        p, t = term2rdfa(
            cur, all_prefixes, treename, predicate_ids, term_id, stanza, title=title, href=href
        )
        ps.update(p)
        body.append(t)

    if not title:
        title = treename + " Browser"

    # Create the prefix element
    pref_strs = []
    for prefix, base in all_prefixes:
        pref_strs.append(f"{prefix}: {base}")
    pref_str = "\n".join(pref_strs)

    body_wrapper = ["div", {"prefix": pref_str}]
    if include_search:
        body_wrapper.append(
            [
                "div",
                {"class": "form-row mt-2 mb-2"},
                [
                    "input",
                    {
                        "id": f"statements-typeahead",
                        "class": "typeahead form-control",
                        "type": "text",
                        "value": "",
                        "placeholder": "Search",
                    },
                ],
            ]
        )
    body = body_wrapper + body

    # JQuery
    body.append(
        [
            "script",
            {
                "src": "https://code.jquery.com/jquery-3.5.1.min.js",
                "integrity": "sha256-9/aliU8dGd2tb6OSsuzixeV4y/faTqgFtohetphbbj0=",
                "crossorigin": "anonymous",
            },
        ]
    )

    if include_search:
        # Add JS imports for running search
        body.append(["script", {"type": "text/javascript", "src": popper_js}])
        body.append(["script", {"type": "text/javascript", "src": bootstrap_js}])
        body.append(["script", {"type": "text/javascript", "src": typeahead_js}])

    # Custom JS for show more children
    js = """function show_children() {
        hidden = $('#children li:hidden').slice(0, 100);
        if (hidden.length > 1) {
            hidden.show();
            setTimeout(show_children, 100);
        } else {
            console.log("DONE");
        }
        $('#more').hide();
    }"""

    # Custom JS for search bar using Typeahead
    if include_search:
        # Built the href to return when you select a term
        href_split = href.split("{curie}")
        before = href_split[0].format(db=treename)
        after = href_split[1].format(db=treename)
        js_funct = f'str.push("{before}" + encodeURIComponent(obj[p]) + "{after}");'

        # Build the href to return names JSON
        remote = "'?text=%QUERY&format=json'"
        if "db=" in href:
            # Add tree name to query params
            remote = f"'?db={treename}&text=%QUERY&format=json'"
        js += (
            """
$('#search-form').submit(function () {
    $(this)
        .find('input[name]')
        .filter(function () {
            return !this.value;
        })
        .prop('name', '');
});
function jump(currentPage) {
  newPage = prompt("Jump to page", currentPage);
  if (newPage) {
    href = window.location.href.replace("page="+currentPage, "page="+newPage);
    window.location.href = href;
  }
}
function configure_typeahead(node) {
  if (!node.id || !node.id.endsWith("-typeahead")) {
    return;
  }
  table = node.id.replace("-typeahead", "");
  var bloodhound = new Bloodhound({
    datumTokenizer: Bloodhound.tokenizers.obj.nonword('short_label', 'label', 'synonym'),
    queryTokenizer: Bloodhound.tokenizers.nonword,
    sorter: function(a, b) {
      return a.order - b.order;
    },
    remote: {
      url: """
            + remote
            + """,
      wildcard: '%QUERY',
      transform : function(response) {
          return bloodhound.sorter(response);
      }
    }
  });
  $(node).typeahead({
    minLength: 0,
    hint: false,
    highlight: true
  }, {
    name: table,
    source: bloodhound,
    display: function(item) {
      if (item.label && item.short_label && item.synonym) {
        return item.short_label + ' - ' + item.label + ' - ' + item.synonym;
      } else if (item.label && item.short_label) {
        return item.short_label + ' - ' + item.label;
      } else if (item.label && item.synonym) {
        return item.label + ' - ' + item.synonym;
      } else if (item.short_label && item.synonym) {
        return item.short_label + ' - ' + item.synonym;
      } else if (item.short_label && !item.label) {
        return item.short_label;
      } else {
        return item.label;
      }
    },
    limit: 40
  });
  $(node).bind('click', function(e) {
    $(node).select();
  });
  $(node).bind('typeahead:select', function(ev, suggestion) {
    $(node).prev().val(suggestion.id);
    go(table, suggestion.id);
  });
  $(node).bind('keypress',function(e) {
    if(e.which == 13) {
      go(table, $('#' + table + '-hidden').val());
    }
  });
}
$('.typeahead').each(function() { configure_typeahead(this); });
function go(table, value) {
  q = {}
  table = table.replace('_all', '');
  q[table] = value
  window.location = query(q);
}
function query(obj) {
  var str = [];
  for (var p in obj)
    if (obj.hasOwnProperty(p)) {
      """
            + js_funct
            + """
    }
  return str.join("&");
}"""
        )

    body.append(["script", {"type": "text/javascript"}, js])

    if standalone:
        # HTML Headers & CSS
        head = [
            "head",
            ["meta", {"charset": "utf-8"}],
            [
                "meta",
                {
                    "name": "viewport",
                    "content": "width=device-width, initial-scale=1, shrink-to-fit=no",
                },
            ],
            ["link", {"rel": "stylesheet", "href": bootstrap_css, "crossorigin": "anonymous"}],
            ["link", {"rel": "stylesheet", "href": "../style.css"}],
            ["title", title],
            [
                "style",
                """
        #annotations {
          padding-left: 1em;
          list-style-type: none !important;
        }
        #annotations ul {
          padding-left: 3em;
          list-style-type: circle !important;
        }
        #annotations ul ul {
          padding-left: 2em;
          list-style-type: none !important;
        }
        .hierarchy {
          padding-left: 0em;
          list-style-type: none !important;
        }
        .hierarchy ul {
          padding-left: 1em;
          list-style-type: none !important;
        }
        .hierarchy ul.multiple-children > li > ul {
          border-left: 1px dotted #ddd;
        }
        .hierarchy .children {
          border-left: none;
          margin-left: 2em;
          text-indent: -1em;
        }
        .hierarchy .children li::before {
          content: "\2022";
          color: #ddd;
          display: inline-block;
          width: 0em;
          margin-left: -1em;
        }
        #nonpeptides .tt-dataset {
          max-height: 300px;
          overflow-y: scroll;
        }
        span.twitter-typeahead .tt-menu {
          cursor: pointer;
        }
        .dropdown-menu, span.twitter-typeahead .tt-menu {
          position: absolute;
          top: 100%;
          left: 0;
          z-index: 1000;
          display: none;
          float: left;
          min-width: 160px;
          padding: 5px 0;
          margin: 2px 0 0;
          font-size: 1rem;
          color: #373a3c;
          text-align: left;
          list-style: none;
          background-color: #fff;
          background-clip: padding-box;
          border: 1px solid rgba(0, 0, 0, 0.15);
          border-radius: 0.25rem; }
        span.twitter-typeahead .tt-suggestion {
          display: block;
          width: 100%;
          padding: 3px 20px;
          clear: both;
          font-weight: normal;
          line-height: 1.5;
          color: #373a3c;
          text-align: inherit;
          white-space: nowrap;
          background: none;
          border: 0; }
        span.twitter-typeahead .tt-suggestion:focus,
        .dropdown-item:hover,
        span.twitter-typeahead .tt-suggestion:hover {
            color: #2b2d2f;
            text-decoration: none;
            background-color: #f5f5f5; }
        span.twitter-typeahead .active.tt-suggestion,
        span.twitter-typeahead .tt-suggestion.tt-cursor,
        span.twitter-typeahead .active.tt-suggestion:focus,
        span.twitter-typeahead .tt-suggestion.tt-cursor:focus,
        span.twitter-typeahead .active.tt-suggestion:hover,
        span.twitter-typeahead .tt-suggestion.tt-cursor:hover {
            color: #fff;
            text-decoration: none;
            background-color: #0275d8;
            outline: 0; }
        span.twitter-typeahead .disabled.tt-suggestion,
        span.twitter-typeahead .disabled.tt-suggestion:focus,
        span.twitter-typeahead .disabled.tt-suggestion:hover {
            color: #818a91; }
        span.twitter-typeahead .disabled.tt-suggestion:focus,
        span.twitter-typeahead .disabled.tt-suggestion:hover {
            text-decoration: none;
            cursor: not-allowed;
            background-color: transparent;
            background-image: none;
            filter: "progid:DXImageTransform.Microsoft.gradient(enabled = false)"; }
        span.twitter-typeahead {
          width: 100%; }
          .input-group span.twitter-typeahead {
            display: block !important; }
            .input-group span.twitter-typeahead .tt-menu {
              top: 2.375rem !important; }""",
            ],
        ]
        body = ["body", {"class": "container"}, body]
        html = ["html", head, body]
    else:
        html = body
    return render(all_prefixes, html, href=href, db=treename)


def create_stanza(results):
    """Create a dictionary version of the results of the stanza query"""
    stanza = []
    for row in results:
        stanza.append(
            {
                "stanza": row[0],
                "subject": row[1],
                "predicate": row[2],
                "object": row[3],
                "value": row[4],
                "datatype": row[5],
                "language": row[6],
            }
        )
    return stanza


def thing2rdfa(cur, all_prefixes, treename, predicate_ids, title=None, href="?id={curie}"):
    """Create a hiccup-style HTML vector for owl:Thing as the parent of all top-level terms."""
    # Select all classes without parents and set them as children of owl:Thing
    cur.execute(
        """SELECT DISTINCT subject FROM statements 
        WHERE subject NOT IN 
            (SELECT subject FROM statements
             WHERE predicate = 'rdfs:subClassOf')
        AND subject IN 
            (SELECT subject FROM statements 
             WHERE predicate = 'rdf:type'
             AND object = 'owl:Class' AND subject NOT LIKE '_:%');"""
    )
    res = cur.fetchall()
    add_children = [x[0] for x in res if x[0] != "owl:Thing"]
    cur.execute(
        """SELECT stanza, subject, predicate, object, value, datatype, language
           FROM statements WHERE stanza = 'owl:Thing'"""
    )
    stanza = cur.fetchall()
    if not stanza:
        stanza = [
            {
                "stanza": "owl:Thing",
                "subject": "owl:Thing",
                "predicate": "rdf:type",
                "object": "owl:Class",
                "value": None,
                "datatype": None,
                "language": None,
            }
        ]
    else:
        stanza = create_stanza(cur.fetchall())
    return term2rdfa(
        cur,
        all_prefixes,
        treename,
        stanza,
        "owl:Thing",
        predicate_ids,
        title=title,
        href=href,
        add_children=add_children,
    )


def curie2iri(prefixes, curie):
    """Convert a CURIE to IRI"""
    if curie.startswith("<"):
        return curie.lstrip("<").rstrip(">")
    for prefix, base in prefixes:
        if curie.startswith(prefix + ":"):
            return curie.replace(prefix + ":", base)
    raise ValueError(f"No matching prefix for {curie}")


def get_entity_type(cur, term_id):
    """Get the OWL entity type for a term."""
    cur.execute(
        f"""SELECT object FROM statements
            WHERE stanza = '{term_id}' AND subject = '{term_id}' AND predicate = 'rdf:type'"""
    )
    res = cur.fetchall()
    if len(res) > 1:
        for r in res:
            if r[0] in TOP_LEVELS:
                return r[0]
        return "owl:Individual"
    elif len(res) == 1:
        entity_type = res[0][0]
        if entity_type == "owl:NamedIndividual":
            entity_type = "owl:Individual"
        return entity_type
    else:
        entity_type = None
        cur.execute(
            f"SELECT predicate FROM statements WHERE stanza = '{term_id}' AND subject = '{term_id}'"
        )
        preds = [row[0] for row in cur.fetchall()]
        if "rdfs:subClassOf" in preds:
            return "owl:Class"
        elif "rdfs:subPropertyOf" in preds:
            return "owl:AnnotationProperty"
        if not entity_type:
            cur.execute(f"SELECT predicate FROM statements WHERE object = '{term_id}'")
            preds = [row[0] for row in cur.fetchall()]
            if "rdfs:subClassOf" in preds:
                return "owl:Class"
            elif "rdfs:subPropertyOf" in preds:
                return "owl:AnnotationProperty"
    return "owl:Class"


def get_hierarchy(cur, term_id, entity_type, add_children=None):
    """Return a hierarchy dictionary for a term and all its ancestors and descendants."""
    # Build the hierarchy
    if entity_type == "owl:Individual":
        cur.execute(
            f"""SELECT DISTINCT object AS parent, subject AS child FROM statements
                WHERE stanza = '{term_id}'
                 AND subject = '{term_id}'
                 AND predicate = 'rdf:type'
                 AND object NOT IN ('owl:Individual', 'owl:NamedIndividual')
                 AND object NOT LIKE '_:%'"""
        )
        res = cur.fetchall()
    else:
        pred = "rdfs:subPropertyOf"
        if entity_type == "owl:Class":
            pred = "rdfs:subClassOf"
        cur.execute(
            f"""WITH RECURSIVE ancestors(parent, child) AS (
                VALUES ('{term_id}', NULL)
                UNION
                -- The children of the given term:
                SELECT object AS parent, subject AS child
                FROM statements
                WHERE predicate = '{pred}'
                  AND object = '{term_id}'
                UNION
                --- Children of the children of the given term
                SELECT object AS parent, subject AS child
                FROM statements
                WHERE object IN (SELECT subject FROM statements
                                 WHERE predicate = '{pred}' AND object = '{term_id}')
                  AND predicate = '{pred}'
                UNION
                -- The non-blank parents of all of the parent terms extracted so far:
                SELECT object AS parent, subject AS child
                FROM statements, ancestors
                WHERE ancestors.parent = statements.stanza
                  AND statements.predicate = '{pred}'
                  AND statements.object NOT LIKE '_:%'
              )
              SELECT * FROM ancestors"""
        )
        res = cur.fetchall()
    if add_children:
        res.extend([[term_id, child] for child in add_children])

    hierarchy = {
        entity_type: {"parents": [], "children": []},
        term_id: {"parents": [], "children": []},
    }
    curies = set()
    for row in res:
        # Consider the parent column of the current row:
        parent = row[0]
        if not parent or parent == "owl:Thing":
            continue
        # If it is not null, add it to the list of all of the compact URIs described by this tree:
        curies.add(parent)
        # If it is not already in the tree, add a new entry for it to the tree:
        if parent not in hierarchy:
            hierarchy[parent] = {
                "parents": [],
                "children": [],
            }

        # Consider the child column of the current row:
        child = row[1]
        if not child:
            continue
        # If it is not null, add it to the list of all the compact URIs described by this tree:
        curies.add(child)
        # If the child is not already in the tree, add a new entry for it to the tree:
        if child not in hierarchy:
            hierarchy[child] = {
                "parents": [],
                "children": [],
            }

        # Fill in the appropriate relationships in the entries for the parent and child:
        hierarchy[parent]["children"].append(child)
        hierarchy[child]["parents"].append(parent)

    if not hierarchy[term_id]["parents"]:
        # Place cur term directly under top level entity
        hierarchy[term_id]["parents"].append(entity_type)
        hierarchy[entity_type]["children"].append(term_id)

    # Add entity type as top level to anything without a parent
    for term_id, mini_tree in hierarchy.items():
        if not mini_tree["parents"]:
            hierarchy[term_id]["parents"].append(entity_type)

    return hierarchy, curies


def get_sorted_predicates(cur, exclude_ids=None):
    """Return a list of predicates IDs sorted by their label, optionally excluding some predicate
    IDs. If the predicate does not have a label, use the ID as the label."""
    exclude = None
    if exclude_ids:
        exclude = ", ".join([f"'{x}'" for x in exclude_ids])

    # Retrieve all predicate IDs
    cur.execute("SELECT DISTINCT predicate FROM statements")
    all_predicate_ids = [x[0] for x in cur.fetchall()]
    if exclude:
        all_predicate_ids = [x for x in all_predicate_ids if x not in exclude_ids]

    # Retrieve predicates with labels
    ap_str = ", ".join([f"'{x}'" for x in all_predicate_ids])
    cur.execute(
        f"""SELECT DISTINCT subject, value
        FROM statements WHERE subject IN ({ap_str}) AND predicate = 'rdfs:label';"""
    )
    predicate_label_map = {x[0]: x[1] for x in cur.fetchall()}

    # Add unlabeled predicates to map with label = ID
    for p in all_predicate_ids:
        if p not in predicate_label_map:
            predicate_label_map[p] = p

    # Return list of keys sorted by value (label)
    return [k for k, v in sorted(predicate_label_map.items(), key=lambda x: x[1].lower())]


def get_ontology(cur, prefixes):
    cur.execute(
        "SELECT subject FROM statements WHERE predicate = 'rdf:type' AND object = 'owl:Ontology'"
    )
    res = cur.fetchone()
    if not res:
        return None, None
    iri = res[0]
    dct = "<http://purl.org/dc/terms/title>"
    for prefix, base in prefixes:
        if base == "http://purl.org/dc/terms/":
            dct = f"{prefix}:title"
    cur.execute(
        f"""SELECT value FROM statements
            WHERE stanza = '{iri}' AND subject = '{iri}' AND predicate = '{dct}'"""
    )
    res = cur.fetchone()
    if not res:
        return iri, None
    return iri, res[0]


def term2rdfa(
    cur,
    prefixes,
    treename,
    predicate_ids,
    term_id,
    stanza,
    title=None,
    add_children=None,
    href="?id={curie}",
):
    """Create a hiccup-style HTML vector for the given term."""
    ontology_iri, ontology_title = get_ontology(cur, prefixes)
    if term_id not in TOP_LEVELS:
        # Get a hierarchy under the entity type
        entity_type = get_entity_type(cur, term_id)
        hierarchy, curies = get_hierarchy(cur, term_id, entity_type, add_children=add_children)
    else:
        # Get the top-level for this entity type
        entity_type = term_id
        if term_id == "ontology":
            hierarchy = {term_id: {"parents": [], "children": []}}
            curies = set()
            if ontology_iri:
                curies.add(ontology_iri)
        else:
            pred = None
            if term_id == "owl:Individual":
                tls = ", ".join([f"'{x}'" for x in TOP_LEVELS.keys()])
                cur.execute(
                    f"""SELECT DISTINCT subject FROM statements
                    WHERE subject NOT IN
                        (SELECT subject FROM statements
                         WHERE predicate = 'rdf:type'
                         AND object NOT IN ('owl:Individual', 'owl:NamedIndividual'))
                    AND subject IN
                        (SELECT subject FROM statements
                         WHERE predicate = 'rdf:type' AND object NOT IN ({tls}))"""
                )
            elif term_id == "rdfs:Datatype":
                cur.execute(
                    """SELECT DISTINCT subject FROM statements
                    WHERE predicate = 'rdf:type' AND object = 'rdfs:Datatype'"""
                )
            else:
                pred = "rdfs:subPropertyOf"
                if term_id == "owl:Class":
                    pred = "rdfs:subClassOf"
                # Select all classes without parents and set them as children of owl:Thing
                cur.execute(
                    f"""SELECT DISTINCT subject FROM statements 
                    WHERE subject NOT IN 
                        (SELECT subject FROM statements
                         WHERE predicate = '{pred}'
                         AND object != 'owl:Thing')
                    AND subject IN 
                        (SELECT subject FROM statements 
                         WHERE predicate = 'rdf:type'
                         AND object = '{term_id}' AND subject NOT LIKE '_:%'
                         AND subject NOT IN ('owl:Thing', 'rdf:type'));"""
                )
            children = [row[0] for row in cur.fetchall()]
            child_children = defaultdict(set)
            if pred and children:
                # Get children of children for classes & properties
                children_str = ", ".join([f"'{x}'" for x in children])
                cur.execute(
                    f"""SELECT DISTINCT object AS parent, subject AS child FROM statements
                    WHERE predicate = '{pred}' AND object IN ({children_str})"""
                )
                for row in cur.fetchall():
                    p = row[0]
                    if p not in child_children:
                        child_children[p] = set()
                    child_children[p].add(row[1])
            hierarchy = {term_id: {"parents": [], "children": children}}
            curies = {term_id}
            for c in children:
                c_children = child_children.get(c, set())
                hierarchy[c] = {"parents": [term_id], "children": list(c_children)}
                curies.update(c_children)
                curies.add(c)

    # Add all of the other compact URIs in the stanza to the set of compact URIs:
    stanza.sort(key=lambda x: x["predicate"])
    for row in stanza:
        curies.add(row.get("subject"))
        curies.add(row.get("predicate"))
        curies.add(row.get("object"))
    curies.discard("")
    curies.discard(None)

    # Get all the prefixes that are referred to by the compact URIs:
    ps = set()
    for curie in curies:
        if not isinstance(curie, str) or len(curie) == 0 or curie[0] in ("_", "<"):
            continue
        prefix, local = curie.split(":")
        ps.add(prefix)

    # Get all of the rdfs:labels corresponding to all of the compact URIs, in the form of a map
    # from compact URIs to labels:
    labels = {}
    ids = "', '".join(curies)
    cur.execute(
        f"""SELECT subject, value
      FROM statements
      WHERE stanza IN ('{ids}')
        AND predicate = 'rdfs:label'
        AND value IS NOT NULL"""
    )
    for row in cur.fetchall():
        labels[row[0]] = row[1]
    for t, o_label in TOP_LEVELS.items():
        labels[t] = o_label
    if ontology_iri and ontology_title:
        labels[ontology_iri] = ontology_title

    obsolete = []
    cur.execute(
        f"""SELECT DISTINCT subject
            FROM statements
            WHERE stanza in ('{ids}')
              AND predicate='owl:deprecated'
              AND value='true'"""
    )
    for row in cur:
        obsolete.append(row[0])

    # If the compact URIs in the labels map are also in the tree, then add the label info to the
    # corresponding node in the tree:
    for key in hierarchy.keys():
        if key in labels:
            hierarchy[key]["label"] = labels[key]

    # Initialise a map with one entry for the tree and one for all of the labels corresponding to
    # all of the compact URIs in the stanza:
    data = {"labels": labels, "obsolete": obsolete, treename: hierarchy, "iri": ontology_iri}

    # Determine the label to use for the given term id when generating RDFa (the term might have
    # multiple labels, in which case we will just choose one and show it everywhere). This defaults
    # to the term id itself, unless there is a label for the term in the stanza corresponding to the
    # label for that term in the labels map:
    if term_id in labels:
        selected_label = labels[term_id]
    else:
        selected_label = term_id
    label = term_id
    for row in stanza:
        predicate = row["predicate"]
        value = row["value"]
        if predicate == "rdfs:label" and value == selected_label:
            label = value
            break

    subject = None
    si = None
    subject_label = None
    if term_id == "ontology" and ontology_iri:
        cur.execute(
            f"""SELECT stanza, subject, predicate, object, value, datatype, language FROM statements
            WHERE subject = '{ontology_iri}'"""
        )
        stanza = create_stanza(cur.fetchall())
        subject = ontology_iri
        subject_label = data["labels"].get(ontology_iri, ontology_iri)
        si = curie2iri(prefixes, subject)
    elif term_id != "ontology":
        subject = term_id
        si = curie2iri(prefixes, subject)
        subject_label = label

    rdfa_tree = term2tree(data, treename, term_id, entity_type, href=href)

    if not title:
        title = treename + " Browser"

    if (term_id in TOP_LEVELS and term_id != "ontology") or (
        term_id == "ontology" and not ontology_iri
    ):
        si = None
        if ontology_iri:
            si = curie2iri(prefixes, ontology_iri)
        items = [
            "ul",
            {"id": "annotations", "class": "col-md"},
            ["p", {"class": "lead"}, "Hello! This is an ontology browser."],
            [
                "p",
                "An ",
                [
                    "a",
                    {"href": "https://en.wikipedia.org/wiki/Ontology_(information_science)"},
                    "ontology",
                ],
                " is a terminology system designed for both humans and machines to read. Click the",
                " links on the left to browse the hierarchy of terms. Terms have parent terms, ",
                "child terms, annotations, and ",
                [
                    "a",
                    {"href": "https://en.wikipedia.org/wiki/Web_Ontology_Language"},
                    "logical axioms",
                ],
                ". The page for each term is also machine-readable using ",
                ["a", {"href": "https://en.wikipedia.org/wiki/RDFa"}, "RDFa"],
                ".",
            ],
        ]
        term = [
            "div",
            ["div", {"class": "row"}, ["h2", title]],
        ]
        if si:
            # If ontology IRI, add it to the page
            term.append(["div", {"class": "row"}, ["a", {"href": si}, si]])
        term.append(["div", {"class": "row", "style": "padding-top: 10px;"}, rdfa_tree, items])
    else:
        items = annotations2rdfa(treename, data, predicate_ids, subject, stanza, href=href)
        term = [
            "div",
            {"resource": subject},
            ["div", {"class": "row"}, ["h2", subject_label]],
            ["div", {"class": "row"}, ["a", {"href": si}, si]],
            ["div", {"class": "row", "style": "padding-top: 10px;"}, rdfa_tree, items],
        ]
    return ps, term


def parent2tree(data, treename, selected_term, selected_children, node, href="?id={curie}"):
    """Return a hiccup-style HTML vector of the full hierarchy for a parent node."""
    cur_hierarchy = ["ul", ["li", tree_label(data, treename, selected_term), selected_children]]
    if node in TOP_LEVELS:
        # Parent is top-level, nothing to add
        return cur_hierarchy

    # Add parents to the hierarchy
    i = 0
    while node and i < 100:
        i += 1
        oc = node
        object_label = tree_label(data, treename, node)
        parents = data[treename][node]["parents"]
        if len(parents) == 0:
            # No parent
            o = [
                "a",
                {"resource": oc, "href": href.format(curie=node, db=treename)},
                object_label,
            ]
            cur_hierarchy = ["ul", ["li", o, cur_hierarchy]]
            break
        parent = parents[0]
        if node == parent:
            # Parent is the same
            o = [
                "a",
                {"resource": oc, "href": href.format(curie=node, db=treename)},
                object_label,
            ]
            cur_hierarchy = ["ul", ["li", o, cur_hierarchy]]
            break
        if parent in TOP_LEVELS:
            href_ele = {"href": href.format(curie=node, db=treename)}
        else:
            href_ele = {
                "about": parent,
                "rev": "rdfs:subClassOf",
                "resource": oc,
                "href": href.format(curie=node, db=treename),
            }
        o = ["a", href_ele, object_label]
        cur_hierarchy = ["ul", ["li", o, cur_hierarchy]]
        node = parent
        if node in TOP_LEVELS:
            break
    return cur_hierarchy


def term2tree(data, treename, term_id, entity_type, href="?id={curie}", max_children=100):
    """Create a hiccup-style HTML hierarchy vector for the given term."""
    if treename not in data or term_id not in data[treename]:
        return ""

    term_tree = data[treename][term_id]
    obsolete = data["obsolete"]
    child_labels = []
    obsolete_child_labels = []
    for child in term_tree["children"]:
        if child in obsolete:
            obsolete_child_labels.append([child, data["labels"].get(child, child)])
        else:
            child_labels.append([child, data["labels"].get(child, child)])
    child_labels.sort(key=lambda x: x[1].lower())
    obsolete_child_labels.sort(key=lambda x: x[1].lower())
    child_labels.extend(obsolete_child_labels)

    if entity_type == "owl:Class":
        predicate = "rdfs:subClassOf"
    elif entity_type == "owl:Individual":
        predicate = "rdf:type"
    else:
        predicate = "rdfs:subPropertyOf"

    # Get the children for our target term
    children = []
    for child, label in child_labels:
        if child not in data[treename]:
            continue
        oc = child
        object_label = tree_label(data, treename, oc)
        o = ["a", {"rev": predicate, "resource": oc}, object_label]
        # Check for children of the child and add a plus next to label if so
        if data[treename][oc]["children"]:
            o.append(PLUS)
        attrs = {}
        if len(children) > max_children:
            attrs["style"] = "display: none"
        children.append(["li", attrs, o])

        if len(children) == max_children:
            total = len(term_tree["children"])
            attrs = {"href": "javascript:show_children()"}
            children.append(["li", {"id": "more"}, ["a", attrs, f"Click to show all {total} ..."]])
    children = ["ul", {"id": "children"}] + children
    if len(children) == 0:
        children = ""
    term_label = tree_label(data, treename, term_id)

    # Get the parents for our target term
    parents = term_tree["parents"]
    if parents:
        hierarchy = ["ul"]
        for p in parents:
            if p.startswith("_:"):
                continue
            hierarchy.append(parent2tree(data, treename, term_id, children.copy(), p, href=href))
    else:
        hierarchy = ["ul", ["li", term_label, children]]

    i = 0
    hierarchies = ["ul", {"id": f"hierarchy", "class": "hierarchy multiple-children col-md"}]
    for t, object_label in TOP_LEVELS.items():
        o = ["a", {"href": href.format(curie=t, db=treename)}, object_label]
        if t == entity_type:
            if term_id == entity_type:
                hierarchies.append(hierarchy)
            else:
                hierarchies.append(["ul", ["li", o, hierarchy]])
            continue
        hierarchies.append(["ul", ["li", o]])
        i += 1
    return hierarchies


def tree_label(data, treename, s):
    """Retrieve the label of a term."""
    node = data[treename][s]
    label = node.get("label", s)
    if s in data["obsolete"]:
        return ["s", label]
    return label


def row2o(_stanza, _data, _uber_row):
    """Given a stanza, a map (`_data`) with entries for the tree structure of the stanza and for all
    of the labels in it, and a row in the stanza, convert the object or value of the row to
    hiccup-style HTML."""

    def renderNonBlank(given_row):
        """Renders the non-blank object from the given row"""
        return [
            "a",
            {"rel": given_row["predicate"], "resource": given_row["object"]},
            _data["labels"].get(given_row["object"], given_row["object"]),
        ]

    def renderLiteral(given_row):
        """Renders the object contained in the given row as a literal IRI"""
        # Literal IRIs are enclosed in angle brackets.
        iri = given_row["object"][1:-1]
        return ["a", {"rel": given_row["predicate"], "href": iri}, iri]

    def getOwlOperands(given_row):
        """Extract all of the operands pointed to by the given row and return them as a list"""
        LOGGER.debug("Finding operands for row with predicate: {}".format(given_row["predicate"]))

        if not given_row["object"].startswith("_:"):
            LOGGER.debug("Found non-blank operand: {}".format(given_row["object"]))
            return [renderNonBlank(given_row)]

        # Find the rows whose subject matches the object from the given row. In general there will
        # be a few. If we find one with an rdf:type predicate then we call the appropriate function
        # to render either a restriction or a class, as the case may be. Otherwise if we find a row
        # with an rdf:first predicate, then if it is a blank node, it points to further operands,
        # which we recursively chase and render, and similarly if the predicate is rdf:rest (which
        # will always have a blank (or nil) object). If the predicate is rdf:first but the object is
        # not blank, then we can render it directly.
        inner_rows = [row for row in _stanza if row["subject"] == given_row["object"]]

        operands = []
        for inner_row in inner_rows:
            inner_subj = inner_row["subject"]
            inner_pred = inner_row["predicate"]
            inner_obj = inner_row["object"]
            LOGGER.debug(f"Found row with <s,p,o> = <{inner_subj}, {inner_pred}, {inner_obj}>")

            if inner_pred == "rdf:type":
                if inner_obj == "owl:Restriction":
                    operands.append(renderOwlRestriction(inner_rows))
                    break
                elif inner_obj == "owl:Class":
                    operands.append(renderOwlClassExpression(inner_rows))
                    break
            elif inner_pred == "rdf:rest":
                if inner_obj != "rdf:nil":
                    operands.append(["span", {"rel": inner_pred}] + getOwlOperands(inner_row))
                else:
                    operands.append(["span", {"rel": inner_pred, "resource": "rdf:nil"}])
                LOGGER.debug(f"Returned from recursing on {inner_pred}")
            elif inner_pred == "rdf:first":
                if inner_obj.startswith("_:"):
                    LOGGER.debug(f"{inner_pred} points to a blank node, following the trail")
                    operands.append(["span", {"rel": inner_pred}] + getOwlOperands(inner_row))
                    LOGGER.debug(f"Returned from recursing on {inner_pred}")
                else:
                    LOGGER.debug(f"Rendering non-blank object with predicate: {inner_pred}")
                    operands.append(renderNonBlank(inner_row))

        return operands

    def renderNaryRelation(class_pred, operands):
        """Render an n-ary relation using the given predicate and operands"""

        def relate_ops(oplist, operator):
            """
            Relate the logical operands in 'oplist' using the given operator word. E.g., if oplist
            contains the logical operands: op1, op2, op3, and the operator is 'and', then an 'and'
            should be rendered in between each of the logical operands.
            """
            # There should always be exactly two operands (see comment below):
            if len(oplist) != 2:
                LOGGER.error(
                    "Unexpected number of operands: {} in relate_ops. Got "
                    "operands: {}".format(len(oplist), oplist)
                )
                return []

            # The list that will be returned, with instances of `operator` inserted:
            related_list = []

            # Get the two operands and their attributes:
            first_op = oplist[0]
            first_op_attrs = first_op[1]
            second_op = oplist[1]
            second_op_attrs = second_op[1]

            # Append the first operand to the related_list:
            related_list.append(first_op)

            # Now handle the second operand:
            if (
                not second_op_attrs.get("rel") == "rdf:rest"
                or second_op_attrs.get("resource") == "rdf:nil"
            ):
                # If there are no more logical operands, append the last rdf:nil span:
                related_list.append(second_op)
            else:
                # Otherwise, logically connect the remaining ones with `operator` and recurse:
                related_list += [" ", operator, " "]
                related_list.append(
                    [second_op[0], second_op[1]] + relate_ops(second_op[2:], operator)
                )

            return related_list

        # There should always be exactly two operands, even if, logically, there are more. Either
        # they'll be literals or they'll be organised into a linked list like the following:
        # [[rdf:first]
        #  [rdf:rest [[rdf:first]
        #             [rdf:rest [rdf:first]
        #                       [rdf:rest ...
        #                                         [rdf:first]
        #                                         [rdf:rest nil]]]...]
        # In this latter case we need to recurse into the rdf:rest spans to find all of the
        # operands other than the first.
        if len(operands) != 2:
            LOGGER.error(
                "Wrong number of operands ({}) to renderNaryRelation. Got class predicate: "
                "{}; operands: {}".format(len(operands), class_pred, operands)
            )
            return ["div"]

        if class_pred == "owl:intersectionOf":
            operator = "and"
        elif class_pred == "owl:unionOf":
            operator = "or"
        else:
            LOGGER.error(f"Unrecognized predicate for n-ary relation: {class_pred}")
            return ["div"]

        owl_div = ["span", {"rel": class_pred}, " ", "("]
        owl_div += relate_ops(operands, operator)
        owl_div.append(")")
        return owl_div

    def renderUnaryRelation(class_pred, operands):
        """Render a unary relation using the given predicate and operands"""
        if len(operands) != 1:
            LOGGER.error(
                f"Something is wrong. Wrong number of operands to '{class_pred}': {operands}"
            )
            return ["div"]

        if class_pred == "owl:complementOf":
            operator = "not"
        elif class_pred == "owl:oneOf":
            operator = "one of"
        else:
            LOGGER.error(f"Unrecognized predicate for unary relation: {class_pred}")
            return ["div"]

        operand = operands[0]
        owl_div = ["span", {"rel": class_pred}, operator, " ", operand]
        return owl_div

    def renderOwlRestriction(given_rows):
        """Renders the OWL restriction described by the given rows"""
        # OWL restrictions are represented using three rows. The first will have the predicate
        # 'rdf:type' and its object should always be 'owl:Restriction'. The second row will have the
        # predicate 'owl:onProperty' and its object will represent the property being restricted,
        # which can be either a blank or a non-blank node. The third row will have either the
        # predicate 'owl:allValuesFrom' or the predicate 'owl:someValuesFrom', which we render,
        # respectively, as 'only' and 'some'. The object of this row is what the property being
        # restricted is being restricted in relation to.
        # E.g., in the restriction: "'has grain' some 'sodium phosphate'": 'has grain' is extracted
        # via the object of the second row, while 'some' and 'sodium phosphate' are
        # extracted via the predicate and object, respectively, of the third row.
        rdf_type_row = [row for row in given_rows if row["predicate"] == "rdf:type"]
        property_row = [row for row in given_rows if row["predicate"] == "owl:onProperty"]
        target_row = [
            row for row in given_rows if row["predicate"] not in ("rdf:type", "owl:onProperty")
        ]
        for rowset in [rdf_type_row, property_row, target_row]:
            if len(rowset) != 1:
                LOGGER.error(f"Rows: {given_rows} do not represent a valid restriction")
                return ["div"]

        property_row = property_row[0]
        target_row = target_row[0]
        rdf_type_row = rdf_type_row[0]
        if rdf_type_row["object"] != "owl:Restriction":
            LOGGER.error(
                "Unexpected rdf:type: '{}' found in OWL restriction".format(rdf_type_row["object"])
            )
            return ["div"]

        target_pred = target_row["predicate"]
        target_obj = target_row["object"]
        LOGGER.debug("Rendering OWL restriction {} for object {}".format(target_pred, target_obj))
        if target_obj.startswith("_:"):
            inner_rows = [row for row in _stanza if row["subject"] == target_obj]
            target_link = renderOwlClassExpression(inner_rows, target_pred)
        else:
            target_link = renderNonBlank(target_row)

        if target_pred == "owl:someValuesFrom":
            operator = "some"
        elif target_pred == "owl:allValuesFrom":
            operator = "only"
        else:
            LOGGER.error("Unrecognised predicate: {}".format(target_pred))
            return ["div"]

        return [
            "span",
            ["span", {"rel": rdf_type_row["predicate"], "resource": rdf_type_row["object"]}],
            [
                "a",
                {"rel": property_row["predicate"], "resource": property_row["object"]},
                _data["labels"].get(property_row["object"], property_row["object"]),
            ],
            " ",
            operator,
            target_link,
        ]

    def renderOwlClassExpression(given_rows, rel=None):
        """Render the OWL class expression pointed to by the given row"""
        # The sub-stanza corresponding to an owl:Class should have two rows. One of these points
        # to the actual class referred to (either a named class or a blank node). From this row we
        # get the subject, predicate, and object to render. The second row will have the object
        # type, which we expect to be 'owl:Class'.
        rdf_type_row = [row for row in given_rows if row["predicate"] == "rdf:type"]
        class_row = [row for row in given_rows if row["predicate"].startswith("owl:")]
        LOGGER.debug(f"Found rows: {rdf_type_row}, {class_row}")

        rdf_type_row = rdf_type_row[0]
        class_row = class_row[0]
        class_subj = class_row["subject"]
        class_pred = class_row["predicate"]
        class_obj = class_row["object"]

        # All blank class expressions will have operands, which we retrieve here:
        operands = getOwlOperands(class_row)

        hiccup = [
            "span",
            ["span", {"rel": rdf_type_row["predicate"], "resource": rdf_type_row["object"]}],
        ]

        # If `rel` is given, insert the attribute into the second position of the hiccup:
        if rel:
            hiccup = hiccup[:1] + [{"rel": rel}] + hiccup[1:]

        LOGGER.debug(f"Rendering <s,p,o> = <{class_subj}, {class_pred}, {class_obj}>")
        if class_pred in ["owl:intersectionOf", "owl:unionOf"]:
            hiccup.append(renderNaryRelation(class_pred, operands))
        elif class_pred in ["owl:complementOf", "owl:oneOf"]:
            hiccup.append(renderUnaryRelation(class_pred, operands))
        elif class_pred == "owl:onProperty":
            hiccup.append(renderOwlRestriction(given_rows))
        elif class_obj.startswith("<"):
            hiccup.append(renderLiteral(class_row))
        else:
            LOGGER.warning(
                f"Rendering for <s,p,o> = <{class_subj}, {class_pred}, {class_obj}> not implemented"
            )
            hiccup.append(["a", {"rel": class_pred}, _data["labels"].get(class_obj, class_obj)])

        return hiccup

    uber_subj = _uber_row["subject"]
    uber_pred = _uber_row["predicate"]
    uber_obj = _uber_row["object"]
    LOGGER.debug(f"Called row2o on <s,p,o> = <{uber_subj}, {uber_pred}, {uber_obj}>")

    if not isinstance(uber_obj, str):
        if _uber_row["value"]:
            LOGGER.debug("Rendering non-string object with value: {}".format(_uber_row["value"]))
            return ["span", {"property": uber_pred}, _uber_row["value"]]
        else:
            LOGGER.error("Received non-string object with null value; returning empty div")
            return ["div"]
    elif uber_obj.startswith("<"):
        LOGGER.debug(f"Rendering literal IRI: {uber_obj}")
        return renderLiteral(_uber_row)
    elif uber_obj.startswith("_:"):
        LOGGER.debug(
            f"Rendering triple with blank object: <s,p,o> = <{uber_subj}, {uber_pred}, {uber_obj}>"
        )
        inner_rows = [row for row in _stanza if row["subject"] == uber_obj]
        object_type = [row for row in inner_rows if row["predicate"] == "rdf:type"]
        if len(object_type) != 1:
            LOGGER.warning(f"Wrong number of object types found for {uber_obj}: {object_type}")
        object_type = object_type[0]["object"] if len(object_type) > 0 else None

        if object_type == "owl:Class":
            LOGGER.debug(f"Rendering OWL class pointed to by {uber_obj}")
            return ["span", {"rel": uber_pred}, renderOwlClassExpression(inner_rows)]
        elif object_type == "owl:Restriction":
            LOGGER.debug(f"Rendering OWL restriction pointed to by {uber_obj}")
            return ["span", {"rel": uber_pred}, renderOwlRestriction(inner_rows)]
        else:
            if not object_type:
                LOGGER.warning(f"Could not determine object type for {uber_pred}")
            else:
                LOGGER.warning(f"Unrecognised object type: {object_type} for predicate {uber_pred}")
            return ["span", {"property": uber_pred}, uber_obj]
    else:
        LOGGER.debug(
            f"Rendering non-blank triple: <s,p,o> = <{uber_subj}, {uber_pred}, {uber_obj}>"
        )
        return renderNonBlank(_uber_row)


if __name__ == "__main__":
    main()
